#!/usr/bin/env python3

from typing import Any, Iterator

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroAuth(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO Auth REST API.
    See https://core.arago.co/help/specs/?url=definitions/auth.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='auth',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations against the auth API
    ###############################################################################################################

    def get_identity(self) -> dict:
        """
        HIRO REST query API: `GET self._auth_endpoint + '/me/account'`

        :return: The result payload
        """
        url = self.endpoint + '/me/account'
        return self.get(url)

    def get_avatar(self) -> Iterator[bytes]:
        """
        HIRO REST query API: `GET self._auth_endpoint + '/me/avatar'`

        :return: The result payload generator over binary data. Complete binary payload is an image/png.
        """
        url = self.endpoint + '/me/avatar'
        yield from self.get_binary(url, accept='image/png')

    def put_avatar(self, data: Any) -> dict:
        """
        HIRO REST query API: `PUT self._auth_endpoint + '/me/avatar'`

        :param data: Binary data for image/png of avatar.
        :return: The result payload
        """
        url = self.endpoint + '/me/avatar'
        return self.put_binary(url, data, content_type='image/png')

    def change_password(self, old_password: str, new_password: str) -> dict:
        """
        HIRO REST query API: `PUT self._auth_endpoint + '/me/password'`

        :param old_password: The old password to replace.
        :param new_password: The new password.
        :return: The result payload
        """
        url = self.endpoint + '/me/password'

        data = {
            "oldPassword": old_password,
            "newPassword": new_password
        }

        return self.put(url, data)

    def get_profile(self) -> dict:
        """
        HIRO REST query API: `GET self._auth_endpoint + '/me/profile`

        :return: The result payload
        """
        url = self.endpoint + '/me/profile'
        return self.get(url)

    def post_profile(self, data: dict) -> dict:
        """
        HIRO REST query API: `POST self._auth_endpoint + '/me/profile`

        :param data: The attributes for the profile.
               See https://core.arago.co/help/specs/?url=definitions/auth.yaml#/[Me]_Identity/post_me_profile
        :return: The result payload
        """
        url = self.endpoint + '/me/profile'
        return self.post(url, data)

    def get_roles(self) -> dict:
        """
        HIRO REST query API: `GET self._auth_endpoint + '/me/roles`

        :return: The result payload
        """
        url = self.endpoint + '/me/roles'
        return self.get(url)

    def get_teams(self) -> dict:
        """
        HIRO REST query API: `GET self._auth_endpoint + '/me/teams'`

        :return: The result payload
        """
        url = self.endpoint + '/me/teams'
        return self.get(url)



import concurrent.futures
import queue
from abc import abstractmethod
from enum import Enum
from typing import Optional, Tuple, Any, Iterator, IO

from requests.exceptions import RequestException

from hiro_graph_client.client import HiroGraph
from hiro_graph_client.clientlib import AbstractTokenApiHandler


class Result(Enum):
    SUCCESS = "success"
    FAILURE = "fail"


class Action(Enum):
    UNDEFINED = "undefined"
    CREATE = "create"
    UPDATE = "update"
    DELETE = "delete"


class Entity(Enum):
    UNDEFINED = "undefined"
    VERTEX = "vertex"
    EDGE = "edge"
    TIMESERIES = "timeseries"
    ATTACHMENT = "attachment"


class HiroResultCallback:
    """
    Abstract class for objects that receive each result of a command in its method *result(...)*.
    Objects of this type are given to the *HiroGraphBatch* as parameter *callback=*.
    """

    @abstractmethod
    def result(self, data: Any, code: int) -> None:
        pass


class AbstractIOCarrier:
    """
    Abstract class that handles IO. When a child of this class is encountered, its IO is opened and read,
    then closed.
    """
    __io_base: IO = None
    """ Private reference to the IO object. Starts with value None and needs to be set in method *start()*. """

    def __enter__(self) -> IO:
        """ To be able to use *with <child of AbstractIOCarrier> as io_item:* """
        return self.open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ To be able to use *with <child of AbstractIOCarrier> as io_item:* """
        self.close()

    @property
    def io_base(self) -> IO:
        """
        Property reader io_base
        :return: Value of *self.__io_base*.
        """
        return self.__io_base

    @io_base.setter
    def io_base(self, io_base: IO) -> None:
        """
        Property setter io_base. This property has to be set in *start()* by the child class.
        """
        self.__io_base = io_base

    @abstractmethod
    def open(self) -> IO:
        """
        Abstract base class for opening IO. Needs to set *self.io_base* to the opened IO object.

        :return: The IO opened
        :raises IOError: Any IO Error
        :raises RuntimeError: When this method is not overwritten by a child class and got called directly.
        """
        raise RuntimeError("Cannot execute method of abstract class.")

    def close(self) -> None:
        """
        Close the IO
        """
        if self.__io_base:
            self.__io_base.close()


class BasicFileIOCarrier(AbstractIOCarrier):
    """
    An IO Carrier for file operations.
    """

    filename: str
    mode: str

    def __init__(self, filename: str, mode: str = 'rb'):
        """
        Constructor

        :param filename: Local filename to start
        :param mode: Open mode. Default is 'rb'.
        """
        self.filename = filename
        self.mode = mode

    def open(self) -> IO:
        """
        Just start the included *self.filename* with *self.mode*.

        :return: The IO handle after start.
        :raises IOError: Any IO error start might raise.
        """
        self.io_base = open(self.filename, self.mode)
        return self.io_base


class SessionData:
    """
    Contains caches and session parameters and the token handler.
    At the moment it carries *xid_cache*, an *edge_store* and a *content_store*. And the *token_handler*.
    """
    xid_cache: dict
    """Cache for xid:id"""

    edge_store: dict
    """Stores a copy of '_edge_data' under the value of 'ogit/_id' as key."""

    timeseries_store: dict
    """Stores a copy of '_timeseries_data' under the value of 'ogit/_id' as key."""

    content_store: dict
    """Stores a copy of '_content_data' under the value of 'ogit/_id' as key."""

    def __init__(self, enable_cache: bool = True):
        """
        Constructor

        :param enable_cache: Enable or disable the xid_cache. Default is True.
        """
        self.xid_cache = {} if enable_cache else None
        self.edge_store = {}
        self.content_store = {}
        self.timeseries_store = {}

    def get_id(self, ogit_xid: str) -> Optional[str]:
        """
        Get an ogit/_id from an ogit/_xid.

        :param ogit_xid: The ogit/_xid to use.
        :return: The ogit/_id from the cache or None if it is not in the cache.
        """
        return None if self.xid_cache is None else self.xid_cache.get(ogit_xid)

    def has_id(self, ogit_id: str) -> bool:
        """
        Check for existence of ogit_id in the cache.

        :param ogit_id: ogit/_id to look for.
        :return: True if ogit_id exists in the cache, False otherwise.
        """
        return False if self.xid_cache is None else bool(list(self.xid_cache.values()).count(ogit_id))

    def register_xid(self, ogit_xid: str, ogit_id: str) -> None:
        """
        Registers a new xid - id mapping in the *xid_cache* unless it or any params are None.

        :param ogit_xid: The xid (key) of the *xid_cache*.
        :param ogit_id: The id (value) of the *xid_cache*.
        """
        if None not in [self.xid_cache, ogit_xid, ogit_id]:
            self.xid_cache[ogit_xid] = ogit_id

    def register_response(self, attributes: dict, response: dict) -> None:
        """
        Collect all data that shall be saved in the session from the attributes and HIRO response.

        Registers a new xid:id mapping in the *xid_cache* unless it or any params are None.

        Saves any edge data, content data and timeseries data from the attributes under the ogit/_id given by the
        response.

        :param attributes: Original attributes from the Runner command
        :param response: Response from the backend after the Runner command ran.
        """
        ogit_id = response.get("ogit/_id")
        ogit_xid = response.get("ogit/_xid")
        edge_data = attributes.get("_edge_data")
        timeseries_data = attributes.get("_timeseries_data")
        content_data = attributes.get("_content_data")

        self.register_xid(ogit_id, ogit_xid)

        if None not in [self.edge_store, ogit_id, edge_data]:
            self.edge_store[ogit_id] = edge_data.copy()

        if None not in [self.timeseries_store, ogit_id, timeseries_data]:
            self.timeseries_store[ogit_id] = timeseries_data.copy()

        if None not in [self.content_store, ogit_id, content_data]:
            self.content_store[ogit_id] = content_data.copy()

    def unregister_by_response(self, response: dict) -> None:
        """
        Remove all cached values using HIRO response data.

        :param response: Response data from a removed vertex.
        """
        ogit_id = response.get("ogit/_id")
        if not ogit_id:
            return

        if self.edge_store:
            self.edge_store.pop(ogit_id, None)

        if self.timeseries_store:
            self.timeseries_store.pop(ogit_id, None)

        if self.content_store:
            self.content_store.pop(ogit_id, None)

        if self.xid_cache:
            for k, v in self.xid_cache.items():
                if v == ogit_id:
                    del self.xid_cache[k]
                    break


class HiroBatchRunner:
    """
    Handles a batch of data intended for the same command.

    This is an abstract class. The real commands need to derive from this.
    """

    entity: Entity
    action: Action

    connection: HiroGraph

    session_data: SessionData

    def __init__(self,
                 entity: Entity,
                 action: Action,
                 session_data: SessionData,
                 connection: HiroGraph):
        """
        Constructor

        :param entity: Entity that gets handled (vertex, edge, timeseries etc.)
        :param action: Action to execute on these entities (create, update, delete)
        :param connection: The hiro client to use.
        :param session_data: Carries session data like caches.
        """
        self.connection = connection
        self.entity = entity
        self.action = action

        self.session_data = session_data

    def get_id_by_xid(self, ogit_xid: str) -> Optional[str]:
        """
        Get an ogit/_id from the *xid_cache* or get it from HiroGraph when it is not in the cache is or it is
        disabled.

        :param ogit_xid: The xid to use to look for an ogit/_id.
        :return: The ogit/_id found or None
        """
        ogit_id = self.session_data.get_id(ogit_xid)

        if ogit_id is None:
            id_result = self.connection.get_node_by_xid(ogit_xid, 'ogit/_id')
            items: list = id_result.get('items')
            if items is None:
                raise KeyError("Response contains no key 'items': \"{}\"".format(id_result))

            if items:
                ogit_id = items[0].get('ogit/_id')

            self.session_data.register_xid(ogit_xid, ogit_id)

        return ogit_id

    def check_id(self, ogit_id: str) -> Optional[str]:
        """
        Check for existence of an id in *xid_cache* or, if not found there or cache is disabled, in HiroGraph.

        :param ogit_id: The ogit/_id to check.
        :return: The ogit/_id of the result or None
        """
        if self.session_data.has_id(ogit_id):
            return ogit_id

        id_result = self.connection.get_node(ogit_id, 'ogit/_id')
        return id_result.get('ogit/_id')

    def get_vertex_id(self,
                      attributes: dict,
                      id_key: str = "ogit/_id",
                      xid_key: str = "ogit/_xid") -> Optional[str]:
        """
        Get the ogit/_id from the source data given in *attributes*.

        * If the ogit/_id is present in *attributes*, return it.
        * If ogit/_id is missing but ogit/_xid is present, use the ogit/_xid to get the ogit/_id from HiroGraph and
          return it.

        :param attributes: Dict of attributes.
        :param id_key: The key for the ogit/_id in *attributes*. Defaults to "ogit/_id".
        :param xid_key: The key for the ogit/_id in *attributes*. Defaults to "ogit/_xid".

        :return: The ogit/_id found or None otherwise.
        """

        ogit_id = attributes.get(id_key)

        if ogit_id is None:
            ogit_xid = attributes.get(xid_key)
            if ogit_xid:
                ogit_id = self.get_id_by_xid(ogit_xid)

        return ogit_id

    def get_and_check_vertex_id(self,
                                attributes: dict,
                                id_key: str = "ogit/_id",
                                xid_key: str = "ogit/_xid") -> str:
        """
        Get the ogit/_id from the source data given in *attributes*.

        * If the ogit/_id is present in *attributes*, return it.
        * If ogit/_id is missing but ogit/_xid is present, use the ogit/_xid to get the ogit/_id from HiroGraph and
          return it.

        :param attributes: Dict of attributes.
        :param id_key: The key for the ogit/_id in *attributes*. Defaults to "ogit/_id".
        :param xid_key: The key for the ogit/_id in *attributes*. Defaults to "ogit/_xid".

        :return: The ogit/_id found.

        :raises ValueError: When no ogit/_id can be found.
        """

        ogit_id = self.get_vertex_id(attributes, id_key, xid_key)

        if ogit_id is None or not ogit_id:
            raise SourceValueError(
                "\"{}\" not found or empty in attributes and cannot be determined by any \"{}\"".format(id_key,
                                                                                                        xid_key))

        return ogit_id

    @staticmethod
    def get_and_check(attributes: dict, key: str, name: str = 'attributes') -> Any:
        """
        Raise ValueError when key is not in attributes or the value behind the key is empty.

        :param attributes: Dict of attributes.
        :param key: The key to look for in *attributes*.
        :param name: Name of the attributes dict. Default is 'attributes'.
        :return: the value in *attributes* of the key.
        :raises ValueError: When 'key' does not exist in *attributes*.
        """
        attribute = attributes.get(key)
        if not attribute:
            raise SourceValueError("\"{}\" not found or empty in \"{}\".".format(key, name))

        return attribute

    @staticmethod
    def for_each_attribute(attributes: dict, *funcs) -> dict:
        """
        Create a copy of *attributes* where a list of functions is applied to each item. Also does not copy any
        attribute whose key is empty or starts with '_'.

        :param attributes: Dict of attributes.
        :param funcs: Set of functions to apply to each element of *attributes*.
        :return: The resulting copy of *attributes*
        """
        result = {}
        for key, value in attributes.items():
            if not key or key[0] == '_':
                continue

            for func in funcs:
                key, value = func(key, value)
                if key is None:
                    break

            if key is not None:
                result[key] = value

        return result

    def resolve_ids(self, key: str, value: str) -> Tuple[str, str]:
        """
        To be used with self.for_each_attribute()

        Try to resolve_ids keys that start with "id:" or "xid:". Try to find the
        ogit/_id of a vertex by using the value for such a key in the graph. Return a tuple of (key, value) with the key
        without its prefix "id:" or "xid:" and the value resolved to a real "ogit/_id".

        :param key: Attribute key
        :param value: Attribute value
        :return: The changed tuple (key, value).
        """
        if key.startswith("xid:"):
            ogit_id = self.get_id_by_xid(value)

            if ogit_id is None or not ogit_id:
                raise ValueError(
                    "Cannot resolve xid \"{}\" of \"{}\".".format(value, key))

            return key[4:], ogit_id

        elif key.startswith("id:"):
            ogit_id = self.check_id(value)

            if ogit_id is None or not ogit_id:
                raise ValueError(
                    "Cannot find id \"{}\" of \"{}\".".format(value, key))

            return key[3:], ogit_id

        else:
            return key, value

    @staticmethod
    def sanitize_for_update(key: str, value: str) -> Tuple[Optional[str], Optional[str]]:
        """
        To be used with self.for_each_attribute()

        Return (None, None) when key starts with "ogit/_" unless "ogit/_owner", "ogit/_content" or "ogit/_tags".

        :param key: Attribute key
        :param value: Attribute value
        :return: The changed tuple (key, value) or (None, None) when this item should be skipped.
        """
        if key.startswith("ogit/_") and key not in ["ogit/_owner", "ogit/_content", "ogit/_tags"]:
            return None, None

        return key, value

    @staticmethod
    def success_message(entity: Entity, action: Action, data: dict) -> dict:
        """
        Success message format

        ::

            {
                "status": "success",
                "entity": entity.value,
                "action": action.value,
                "data": data
            }

        :param entity: Entity handled
        :param action: Action done
        :param data: JSON to return
        :return: The message
        """
        return {
            "status": Result.SUCCESS.value,
            "entity": entity.value,
            "action": action.value,
            "data": data
        }

    @staticmethod
    def error_message(entity: Entity,
                      action: Action,
                      error: Exception,
                      original: dict,
                      status_code: int = None) -> dict:
        """
        Failure message format

        ::

            {
                "status": "fail",
                "entity": entity.value,
                "action": action.value,
                "data": {
                    "error": error.__class__.__name__,
                    "message": str(error),
                    "original_data": original
                }
            }

        :param entity: Entity handled
        :param action: Action done
        :param error: The exception raised
        :param original: The data that lead to the exception
        :param status_code: HTTP status code if available
        :return: The message
        """
        return {
            "status": Result.FAILURE.value,
            "entity": entity.value,
            "action": action.value,
            "data": {
                "error": error.__class__.__name__,
                "code": status_code,
                "message": str(error),
                "original_data": original
            }
        }

    def run(self, attributes: dict, result_queue: queue.Queue) -> None:
        """
        Run the Command with all data given by *attributes*.

        This is the enclosing code for all batch runners. The implementation of a handling a single entry of
        the *attributes* is defined in derived ...Runner-classes.

        :param attributes: Dict with attributes to handle in HIRO.
        :param result_queue: Queue receiving the results.
        """
        try:
            response: dict = self.run_item(attributes)

            response_code = 200
            message = self.success_message(self.entity, self.action, response)

        except RequestException as error:
            response_code = error.response.status_code if error.response is not None else 999
            message = self.error_message(self.entity,
                                         self.action,
                                         error,
                                         attributes,
                                         response_code)

        except SourceValueError as error:
            response_code = 400
            message = self.error_message(self.entity, self.action, error, attributes, 400)

        except Exception as error:
            response_code = 500
            message = self.error_message(self.entity, self.action, error, attributes, 500)

        result_queue.put((message, response_code))

    @abstractmethod
    def run_item(self, attributes: dict) -> dict:
        """
        Abstract method overwritten by derived runner classes.

        :param attributes: A dict of attributes to handle.
        :return: A response dict - usually directly the structure received from the backend.
        """
        raise RuntimeError("Cannot _run within HiroCommandBatch directly.")


class CreateVerticesRunner(HiroBatchRunner):
    """
    Create vertices
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Create vertices

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.VERTEX, Action.CREATE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to create a vertex from.
        :return: A response dict - usually directly the structure received from the backend.
        """

        ogit_type = self.get_and_check(attributes, "ogit/_type")
        final_attributes = self.for_each_attribute(attributes, self.resolve_ids)

        response: dict = self.connection.create_node(final_attributes, ogit_type)

        self.session_data.register_response(attributes, response)

        return response


class UpdateVerticesRunner(HiroBatchRunner):
    """
    Update vertices
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Update vertices

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.VERTEX, Action.UPDATE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to update a vertex from.
        :return: A response dict - usually directly the structure received from the backend.
        """
        ogit_id = self.get_and_check_vertex_id(attributes)
        final_attributes = self.for_each_attribute(attributes, self.resolve_ids, self.sanitize_for_update)

        response: dict = self.connection.update_node(ogit_id, final_attributes)

        self.session_data.register_response(attributes, response)

        return response


class DeleteVerticesRunner(HiroBatchRunner):
    """
    Delete vertices
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Delete vertices

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.VERTEX, Action.DELETE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to delete a vertex from.
        :return: A response dict - usually directly the structure received from the backend.
        """
        ogit_id = self.get_and_check_vertex_id(attributes)

        response: dict = self.connection.delete_node(ogit_id)

        self.session_data.unregister_by_response(response)

        return response


class HandleVerticesRunner(HiroBatchRunner):
    """
    Handle vertices. Either update or create them based on incoming payload entries.
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Handle vertices. Either update or create them based on incoming payload entries.

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.VERTEX, Action.UNDEFINED, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to handle a vertex from. Updates the vertex when it can be found via
                           ogit/_id or ogit/_xid, creates the vertex otherwise when ogit/_type is present.
        :return: A response dict - usually directly the structure received from the backend.
        """
        self.action = Action.UNDEFINED

        ogit_id = self.get_vertex_id(attributes)

        self.action = Action.UPDATE if ogit_id else Action.CREATE

        if self.action == Action.CREATE:
            ogit_type = self.get_and_check(attributes, "ogit/_type")
            final_attributes = self.for_each_attribute(attributes, self.resolve_ids)

            response: dict = self.connection.create_node(final_attributes, ogit_type)
        else:
            final_attributes = self.for_each_attribute(attributes, self.resolve_ids, self.sanitize_for_update)

            response: dict = self.connection.update_node(ogit_id, final_attributes)

        self.session_data.register_response(attributes, response)

        return response


class CreateEdgesRunner(HiroBatchRunner):
    """
    Create edges between vertices
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Create edges between vertices

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.EDGE, Action.CREATE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to create edges from.
        :return: A response dict - usually directly the structure received from the backend.
        """
        from_node_id = self.get_and_check_vertex_id(attributes, "from:ogit/_id", "from:ogit/_xid")
        to_node_id = self.get_and_check_vertex_id(attributes, "to:ogit/_id", "to:ogit/_xid")
        verb = self.get_and_check(attributes, "verb")

        return self.connection.connect_nodes(from_node_id, verb, to_node_id)


class DeleteEdgesRunner(HiroBatchRunner):
    """
    Delete edges between vertices
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Delete edges between vertices

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.EDGE, Action.DELETE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to delete edges from.
        :return: A response dict - usually directly the structure received from the backend.
        """
        from_node_id = self.get_and_check_vertex_id(attributes, "from:ogit/_id", "from:ogit/_xid")
        to_node_id = self.get_and_check_vertex_id(attributes, "to:ogit/_id", "to:ogit/_xid")
        verb = self.get_and_check(attributes, "verb")

        return self.connection.disconnect_nodes(from_node_id, verb, to_node_id)


class AddTimeseriesRunner(HiroBatchRunner):
    """
    Attach timeseries values to a vertex.
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Attach timeseries values to a vertex.

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.TIMESERIES, Action.CREATE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to attach timeseries to a vertex.
        :return: A response dict - usually directly the structure received from the backend.
        """
        node_id = self.get_and_check_vertex_id(attributes)
        items = self.get_and_check(attributes, "items")

        return self.connection.post_timeseries(node_id, items)


class AddAttachmentRunner(HiroBatchRunner):
    """
    Attach an attachment to a vertex.
    """

    def __init__(self, session_data: SessionData, connection: HiroGraph):
        """
        Attach an attachment to a vertex.

        :param session_data: Required: Session data / caches.
        :param connection: Required: The handler for the connection to HIRO HiroGraph.
        """
        super().__init__(Entity.ATTACHMENT, Action.CREATE, session_data, connection)

    def run_item(self, attributes: dict) -> dict:
        """
        :param attributes: A dict of attributes to attach attachments to a vertex.
        :return: A response dict - usually directly the structure received from the backend.
        """
        node_id = self.get_and_check_vertex_id(attributes)
        content_data = self.get_and_check(attributes, '_content_data')
        mimetype = content_data.get('mimetype')

        data = content_data.get('data')

        if isinstance(data, AbstractIOCarrier):
            with data as io_item:
                return self.connection.post_attachment(node_id=node_id,
                                                       data=io_item,
                                                       content_type=mimetype)
        elif data:
            return self.connection.post_attachment(node_id=node_id,
                                                   data=data,
                                                   content_type=mimetype)
        else:
            raise SourceValueError('"data" not found or empty in "attributes._content_data".')


class HiroGraphBatch:
    """
    This class handles lists of vertex-, edge- or timeseries-data via HiroGraph.
    """
    request_queue: queue.Queue
    result_queue: queue.Queue

    callback: HiroResultCallback

    _api_handler: AbstractTokenApiHandler

    use_xid_cache: bool
    """Use xid caching. Default is True when omitted or set to None."""

    commands = [
        "create_vertices",
        "update_vertices",
        "handle_vertices",
        "handle_vertices_combined",
        "delete_vertices",
        "create_edges",
        "delete_edges",
        "add_timeseries",
        "add_attachments"
    ]
    """This is the list of commands (method names) that HiroGraphBatch handles."""

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 callback: HiroResultCallback = None,
                 use_xid_cache: bool = True,
                 max_parallel_workers: int = 8,
                 queue_depth: int = None):
        """
        Constructor

        Use the connection to HIRO HiroGraph either by giving a predefined *hiro_token* or by
        specifying all other parameters needed for authentication.

        :param api_handler: External API handler.
        :param callback: required when multi_command() is used, optional otherwise: Callback object for results.
        :param use_xid_cache: Use xid caching. Default is True when omitted or set to None.
        :param max_parallel_workers: Amount of maximum parallel workers for requests. Default is 8.
        :param queue_depth: Amount of entries the *self.request_queue* and *self.result_queue* can hold. Default is to
                            set it to the same value as *parallel_workers*.
        """
        if not api_handler:
            raise ValueError('Need attribute "api_handler" for HIRO Graph API.')

        self._api_handler = api_handler

        self.request_queue = queue.Queue(maxsize=queue_depth or max_parallel_workers)
        self.result_queue = queue.Queue(maxsize=queue_depth or max_parallel_workers)

        self.callback = callback

        self.max_parallel_workers = max_parallel_workers

        self.use_xid_cache = False if use_xid_cache is False else True

    def __prepare_session_and_connection(self,
                                         session: SessionData,
                                         connection: HiroGraph) -> Tuple[SessionData, HiroGraph]:
        """
        Initialize session and connection if either of them is unset.

        :param session: Predefined session - if any.
        :param connection:  Predefined connection - if any.
        :return: A tuple of (SessionData, HiroGraph)
        """
        if not session:
            session = SessionData(self.use_xid_cache)
        if not connection:
            connection = HiroGraph(api_handler=self._api_handler)
        return session, connection

    def create_vertices(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Create vertex from *attributes*.

        :param attributes: Dict containing the attributes for the vertex.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return CreateVerticesRunner(session, connection).run(attributes, self.result_queue)

    def update_vertices(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Update vertex from *attributes*.

        Attributes needs at least a key "ogit/_id" or "ogit/_xid" to find the vertex to be updated.
        It then sanitizes the payload by ignoring every attribute starting with "ogit/_" (unless "ogit/_owner",
        "ogit/_content" or "ogit/_tags") of the dict before attempting to update.

        :param attributes: Dict containing the attributes for the vertex.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return UpdateVerticesRunner(session, connection).run(attributes, self.result_queue)

    def delete_vertices(self, attributes: dict, connection: HiroGraph, session: SessionData = None):
        """
        Delete vertex given by *attributes*.

        Attributes needs at least a key "ogit/_id" or "ogit/_xid" to find the vertex to be deleted.

        :param attributes: Dict containing the attributes for the vertex.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return DeleteVerticesRunner(session, connection).run(attributes, self.result_queue)

    def handle_vertices(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Handles vertex from *attributes*. Tries to update or create.

        When attributes contain "ogit/_id" or "ogit/_xid", they will be used to update an
        existing vertex, otherwise the attributes will be used to create a new vertex.

        :param attributes: Dict containing the attributes for the vertex.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return HandleVerticesRunner(session, connection).run(attributes, self.result_queue)

    def create_edges(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Create edge from *attributes*.

        Attributes needs a key "from:ogit/_id" or "from:ogit/_xid" and
        "to:ogit/_id" or "to:ogit/_xid" as well as "verb" to be able to determine the vertices to connect.

        :param attributes: Dict containing the fields "from:...,verb,to:..." for the edge.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return CreateEdgesRunner(session, connection).run(attributes, self.result_queue)

    def delete_edges(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Delete edge given by *attributes*.

        Attributes needs a key "from:ogit/_id" or "from:ogit/_xid" and
        "to:ogit/_id" or "to:ogit/_xid" as well as "verb" to be able to determine the edge between the vertices to
        delete.

        :param attributes: Dict containing the "from:...,verb,to:..." fields for the edge.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return DeleteEdgesRunner(session, connection).run(attributes, self.result_queue)

    def add_timeseries(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Update vertices with timeseries data.

        Attributes needs at least a key "ogit/_id" or "ogit/_xid" to find the vertex to be updated.

        :param attributes: Contains the timeseries items.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return AddTimeseriesRunner(session, connection).run(attributes, self.result_queue)

    def add_attachments(self, attributes: dict, connection: HiroGraph = None, session: SessionData = None):
        """
        Add attachment to vertex.

        Attributes needs at least a key "ogit/_id" or "ogit/_xid" to find the vertex to be updated, and a dict
        '_content_data' containing the payload to update under key 'data' - use an IO class of AbstractIOCarrier
        for streaming.

        Example:
        ::

            {
                "ogit/_id": "",
                "_content_data": {
                    "data": "(payload)",
                    "mimetype": "(content-type)"
                }
            }

        :param attributes: Contains the attachment data.
        :param connection: optional: Connection to use. A new connection will be used if this is not set.
        :param session: optional: Persistent data for the current session. Use a local session if this is not set.
        """
        session, connection = self.__prepare_session_and_connection(session, connection)
        return AddAttachmentRunner(session, connection).run(attributes, self.result_queue)

    def _edges_from_session(self, session: SessionData) -> None:
        """
        Recreate attributes to create edges saved in a session.

        :param session: The session with the edge data.
        """
        for ogit_id, edge_data in session.edge_store.items():
            for edge in edge_data:
                other_ogit_id = edge.get("vertex_id")
                other_ogit_xid = edge.get("vertex_xid")
                verb = edge.get('verb')
                direction = edge.get("direction")

                attributes = {
                    "verb": verb
                }
                if direction == 'in':
                    attributes["to:ogit/_id"] = ogit_id
                    if other_ogit_id:
                        attributes["from:ogit/_id"] = other_ogit_id
                    else:
                        attributes["from:ogit/_xid"] = other_ogit_xid
                else:
                    attributes["from:ogit/_id"] = ogit_id
                    if other_ogit_id:
                        attributes["to:ogit/_id"] = other_ogit_id
                    else:
                        attributes["to:ogit/_xid"] = other_ogit_xid

                self.request_queue.put(('create_edges', attributes))

    def _timeseries_from_session(self, session: SessionData) -> None:
        """
        Recreate attributes to create timeseries saved in a session.

        :param session: The session with the timeseries data.
        """
        for ogit_id, timeseries_data in session.timeseries_store.items():
            attributes = {
                'ogit/_id': ogit_id,
                'items': timeseries_data
            }

            self.request_queue.put(('add_timeseries', attributes))

    def _attachments_from_session(self, session: SessionData) -> None:
        """
        Recreate attributes to create attachments saved in a session.

        :param session: The session with the attachment data.
        """
        for ogit_id, content_data in session.content_store.items():
            attributes = {
                'ogit/_id': ogit_id,
                '_content_data': content_data
            }

            self.request_queue.put(('add_attachments', attributes))

    def _reader(self, collected_results: list) -> None:
        """
        Thread executor function. Read items from the *self.result_queue* and call the callback function with them.

        Thread exits when *self.result_queue.get()* reads None.
        """
        for result, code in iter(self.result_queue.get, None):
            if self.callback is not None:
                self.callback.result(result, code)
            if collected_results is not None:
                collected_results.append(result)
            self.result_queue.task_done()

    def _worker(self, session: SessionData) -> None:
        """
        Thread executor function. Create a connection, then read data from the *self.request_queue* and execute
        the command with the attributes and session the data from the queue provides.

        Thread exits when *self.request_queue.get()* reads None.

        :param session: The session object to share between all connections.
        """

        connection = HiroGraph(self._api_handler)

        for command, attributes in iter(self.request_queue.get, None):
            func = getattr(self, command, None)
            if func:
                func(attributes, connection, session)
            self.request_queue.task_done()

    def multi_command(self, command_iter: Iterator[dict]) -> Optional[list]:
        """
        Run a multi-command batch.

        The command_iter iterates over a list of dicts with pairs like

        ::

            {
                "[command]": { "[key]": "[value]", ... }
            },
            {
                "[command]": { "[key]": "[value]", ... }
            },
            ...

        or

        ::

            {
                "[command]": [
                    { "[key]": "[value]", ... },
                    { "[key]": "[value]", ... }
                ]
            },
            ...

        with payload being a list of dict containing the attributes to _run with that command.

        :param command_iter: An iterator for a dict of pairs "[command]:payload".
        :return a list with results when no callback is set, None otherwise.
        """
        with concurrent.futures.ThreadPoolExecutor() as executor:

            def _request_queue_put(_command: str,
                                   _attributes: dict,
                                   _parallel_workers: int) -> int:

                if not isinstance(_attributes, dict):
                    raise SourceValueError("Found attributes that are not a dict.")

                if _parallel_workers < self.max_parallel_workers:
                    executor.submit(HiroGraphBatch._worker, self, session)
                    _parallel_workers += 1

                self.request_queue.put((_command, _attributes))
                return _parallel_workers

            session = SessionData(self.use_xid_cache)

            collected_results = [] if self.callback is None else None

            parallel_workers = 0

            executor.submit(HiroGraphBatch._reader, self, collected_results)

            handle_session_data = False
            for command_entry in command_iter:
                for command, attributes in command_entry.items():

                    if command == "handle_vertices_combined":
                        command = "handle_vertices"
                        handle_session_data = True

                    try:
                        if command in self.commands:
                            if isinstance(attributes, list):
                                for attribute_entry in attributes:
                                    parallel_workers = _request_queue_put(
                                        command,
                                        attribute_entry,
                                        parallel_workers)
                            else:
                                parallel_workers = _request_queue_put(
                                    command,
                                    attributes,
                                    parallel_workers)
                        else:
                            raise SourceValueError("No such command \"{}\".".format(command))
                    except SourceValueError as err:
                        sub_result, sub_code = HiroBatchRunner.error_message(
                            Entity.UNDEFINED,
                            Action.UNDEFINED,
                            err,
                            attributes,
                            400), 400

                        self.result_queue.put((sub_result, sub_code))

            if handle_session_data:
                self.request_queue.join()
                self._edges_from_session(session)
                self._timeseries_from_session(session)
                self._attachments_from_session(session)

            self.request_queue.join()
            self.result_queue.join()

            for _ in range(parallel_workers):
                self.request_queue.put(None)

            self.result_queue.put(None)

            return collected_results


class SourceValueError(ValueError):
    """
    An error occurred with missing or invalid source data.
    """
    pass



#!/usr/bin/env python3

import json
import logging
import os
import threading
import time
import urllib
from abc import abstractmethod
from urllib.parse import quote, urlencode

import backoff
import requests
import requests.packages.urllib3.exceptions
from typing import Optional, Any, Iterator, Union, Tuple

from hiro_graph_client.version import __version__

BACKOFF_ARGS = [
    backoff.expo,
    requests.exceptions.RequestException
]
BACKOFF_KWARGS = {
    'max_tries': 2,
    'jitter': backoff.random_jitter,
    'giveup': lambda e: e.response is not None and e.response.status_code < 500
}

logger = logging.getLogger(__name__)
""" The logger for this module """


def accept_all_certs():
    """
    Globally disable InsecureRequestWarning
    """
    AbstractAPI.accept_all_certs = True

    requests.packages.urllib3.disable_warnings(requests.packages.urllib3.exceptions.InsecureRequestWarning)


###################################################################################################################
# Root classes for API
###################################################################################################################

class AbstractAPI:
    """
    This abstract root class contains the methods for HTTP requests used by all API classes. Also contains several
    tool methods for handling headers, url query parts and response error checking.
    """

    accept_all_certs: bool = False

    _client_name: str = "python-hiro-client"

    def __init__(self,
                 root_url: str,
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = 600,
                 client_name: str = None):
        """
        Constructor

        :param root_url: Root uri of the HIRO API, like *https://core.arago.co*.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
                            is given with another value for "User-Agent". Default is "python-hiro-client".
        """

        if not root_url:
            raise ValueError("'root_url' must not be empty.")

        self._root_url = root_url
        self._proxies = proxies
        self._raise_exceptions = raise_exceptions
        self._timeout = timeout

        if client_name:
            self._client_name = client_name

        self._headers = {
            'Content-Type': 'application/json',
            'Accept': 'text/plain, application/json',
            'User-Agent': f"{self._client_name} {__version__}"
        }

        if headers:
            self._headers.update({self._capitalize_header(k): v for k, v in headers.items()})

    def get_root_url(self):
        return self._root_url

    @property
    def user_agent(self):
        return self._headers.get('User-Agent') or self._client_name

    @staticmethod
    def _capitalize_header(name: str) -> str:
        return "-".join([n.capitalize() for n in name.split('-')])

    ###############################################################################################################
    # Basic requests
    ###############################################################################################################

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def get_binary(self, url: str, accept: str = None) -> Iterator[bytes]:
        """
        Implementation of GET for binary data.

        :param url: Url to use
        :param accept: Mimetype for accept. Will be set to */* if not given.
        :return: Yields over raw chunks of the response payload.
        """
        with requests.get(url,
                          headers=self._get_headers(
                              {"Content-Type": None, "Accept": (accept or "*/*")}
                          ),
                          verify=False,
                          timeout=self._timeout,
                          stream=True,
                          proxies=self._get_proxies()) as res:
            self._log_communication(res, response_body=False)
            self._check_response(res)
            self._check_status_error(res)

            yield from res.iter_content(chunk_size=65536)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def post_binary(self, url: str, data: Any, content_type: str = None) -> dict:
        """
        Implementation of POST for binary data.

        :param url: Url to use
        :param data: The payload to POST. This can be anything 'requests.post(data=...)' supports.
        :param content_type: The content type of the data. Defaults to "application/octet-stream" internally if unset.
        :return: The payload of the response
        """
        res = requests.post(url,
                            data=data,
                            headers=self._get_headers(
                                {"Content-Type": (content_type or "application/octet-stream")}
                            ),
                            verify=False,
                            timeout=self._timeout,
                            proxies=self._get_proxies())
        self._log_communication(res, request_body=False)
        return self._parse_json_response(res)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def put_binary(self, url: str, data: Any, content_type: str = None) -> dict:
        """
        Implementation of PUT for binary data.

        :param url: Url to use
        :param data: The payload to PUT. This can be anything 'requests.put(data=...)' supports.
        :param content_type: The content type of the data. Defaults to "application/octet-stream" internally if unset.
        :return: The payload of the response
        """
        res = requests.put(url,
                           data=data,
                           headers=self._get_headers(
                               {"Content-Type": (content_type or "application/octet-stream")}
                           ),
                           verify=False,
                           timeout=self._timeout,
                           proxies=self._get_proxies())
        self._log_communication(res, request_body=False)
        return self._parse_json_response(res)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def get(self, url: str) -> dict:
        """
        Implementation of GET

        :param url: Url to use
        :return: The payload of the response
        """
        res = requests.get(url,
                           headers=self._get_headers({"Content-Type": None}),
                           verify=False,
                           timeout=self._timeout,
                           proxies=self._get_proxies())
        self._log_communication(res)
        return self._parse_json_response(res)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def post(self, url: str, data: Any) -> dict:
        """
        Implementation of POST

        :param url: Url to use
        :param data: The payload to POST
        :return: The payload of the response
        """
        res = requests.post(url,
                            json=data,
                            headers=self._get_headers(),
                            verify=False,
                            timeout=self._timeout,
                            proxies=self._get_proxies())
        self._log_communication(res)
        return self._parse_json_response(res)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def put(self, url: str, data: Any) -> dict:
        """
        Implementation of PUT

        :param url: Url to use
        :param data: The payload to PUT
        :return: The payload of the response
        """
        res = requests.put(url,
                           json=data,
                           headers=self._get_headers(),
                           verify=False,
                           timeout=self._timeout,
                           proxies=self._get_proxies())
        self._log_communication(res)
        return self._parse_json_response(res)

    @backoff.on_exception(*BACKOFF_ARGS, **BACKOFF_KWARGS)
    def delete(self, url: str) -> dict:
        """
        Implementation of DELETE

        :param url: Url to use
        :return: The payload of the response
        """
        res = requests.delete(url,
                              headers=self._get_headers({"Content-Type": None}),
                              verify=False,
                              timeout=self._timeout,
                              proxies=self._get_proxies())
        self._log_communication(res)
        return self._parse_json_response(res)

    ###############################################################################################################
    # Tool methods for requests
    ###############################################################################################################

    def _get_proxies(self) -> dict:
        """
        Create a copy of proxies if they exists or return None

        :return: copy of self._proxies or None
        """
        return self._proxies.copy() if self._proxies else None

    def _get_headers(self, override: dict = None) -> dict:
        """
        Create a header dict for requests. Uses abstract method *self._handle_token()*.

        :param override: Dict of headers that override the internal headers. If a header key is set to value None,
               it will be removed from the headers.
        :return: A dict containing header values for requests.
        """
        headers = self._headers.copy()

        if isinstance(override, dict):
            headers.update({self._capitalize_header(k): v for k, v in override.items()})
            headers = {k: v for k, v in headers.items() if v is not None}

        token = self._handle_token()
        if token:
            headers['Authorization'] = "Bearer " + token

        return headers

    @staticmethod
    def _get_query_part(params: dict) -> str:
        """
        Create the query part of an url. Keys in *params* whose values are set to None are removed.

        :param params: A dict of params to use for the query.
        :return: The query part of an url with a leading '?', or an empty string when query is empty.
        """
        params_cleaned = {k: v for k, v in params.items() if v is not None}
        return ('?' + urlencode(params_cleaned, quote_via=quote, safe="/,")) if params_cleaned else ""

    def _parse_json_response(self, res: requests.Response) -> dict:
        """
        Parse the response of the backend.

        :param res: The result payload
        :return: The result payload
        :raises RequestException: On HTTP errors.
        """
        try:
            self._check_response(res)
            self._check_status_error(res)
            return res.json()
        except (json.JSONDecodeError, ValueError):
            return {"error": {"message": res.text, "code": 999}}

    def _log_communication(self, res: requests.Response, request_body: bool = True, response_body: bool = True) -> None:
        """
        Log communication that flows across requests' methods. Contains options to disable logging of the body portions
        of the requests to avoid dumping large amounts of data and breaking streaming of results.

        This log does not log an exact binary representation of the transmitted bodies but uses the encoding defined
        in *res* to print it as strings.

        Authorization and cookie headers will be obscured by only displaying the last six characters of their values.

        :param res: The response of a request. Also contains the request.
        :param request_body: Option to disable the logging of the request_body.
        :param response_body: Option to disable the logging of the response_body.
        :return:
        """

        def _log_headers(headers) -> str:
            result: str = ""
            for k, v in headers.items():
                cap_key: str = self._capitalize_header(k)
                if cap_key == "Authorization" or cap_key.find("Cookie") != -1:
                    v = f"{v[:6]}[{len(v) - 12} characters hidden]{v[-6:]}"
                result += f"{k}: {v}\n"
            return result

        def _body_str(body: Union[str, bytes], encoding: str) -> str:
            if body is None:
                return ""
            if isinstance(body, bytes):
                body = str(body, encoding or 'utf8')
            return body

        if logger.isEnabledFor(logging.DEBUG):
            log_message = f'''
################ request ################
{res.request.method} {res.request.url}
{_log_headers(res.request.headers)}
{_body_str(res.request.body, res.encoding) if request_body else "(body hidden)"}
################ response ################
{res.status_code} {res.reason} {res.url}
{_log_headers(res.headers)}
{_body_str(res.text, res.encoding) if response_body else "(body hidden)"}
'''

            if not res.ok:
                logger.error(log_message)
            else:
                logger.debug(log_message)

    def _check_status_error(self, res: requests.Response) -> None:
        """
        Catch exceptions and rethrow them with additional information returned by the error response body.

        :param res: The response
        :raises requests.exceptions.HTTPError: When an HTTPError occurred.
        """
        try:
            if self._raise_exceptions:
                res.raise_for_status()
                if res.status_code > 600:
                    raise requests.exceptions.HTTPError(
                        u'%s Illegal return code: %s for url: %s' % (res.status_code, res.reason, res.url),
                        response=res)

        except requests.exceptions.HTTPError as err:
            http_error_msg = str(err.args[0])

            if res.content:
                try:
                    json_result: dict = res.json()
                    message = json_result['error']['message']
                    http_error_msg += ": " + message
                except (json.JSONDecodeError, KeyError):
                    if '_TOKEN' not in res.text:
                        http_error_msg += ": " + str(res.text)

            raise requests.exceptions.HTTPError(http_error_msg, response=err.response) from err

    ###############################################################################################################
    # Response and token handling
    # Child classes have to override those classes for special handling like token and header handling.
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        Root method. No response checking here.

        :param res: The result payload
        """
        return

    def _handle_token(self) -> Optional[str]:
        """
        Just return None, therefore a header without Authorization
        will be created in *self._get_headers()*.

        Does *not* try to obtain or refresh a token.

        :return: Always None here, derived classes should return a token string.
        """
        return None


###################################################################################################################
# TokenApiHandler classes
###################################################################################################################

class AbstractTokenApiHandler(AbstractAPI):
    """
    Root class for all TokenApiHandler classes. This class also handles resolving the current api endpoints.
    """

    def __init__(self,
                 root_url: str,
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = 600,
                 client_name: str = None,
                 custom_endpoints: dict = None):
        """
        Constructor

        Example for custom_endpoints (see params below):

        ::

           {
               "graph": "/api/graph/7.2",
               "auth": "/api/auth/6.2",
               "action-ws": ("/api/action-ws/1.0", "action-1.0.0")
           }


        :param root_url: Root url for HIRO, like https://core.arago.co.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
                            is given with another value for "User-Agent". Default is "python-hiro-client".
        :param custom_endpoints: Optional map of {name:endpoint_path, ...} that overrides or adds to the endpoints taken
               from /api/version. Example see above.
        """
        super().__init__(root_url=root_url,
                         raise_exceptions=raise_exceptions,
                         proxies=proxies,
                         timeout=timeout,
                         headers=headers,
                         client_name=client_name)

        self._version_info = None
        self.custom_endpoints = custom_endpoints

    @staticmethod
    def _remove_slash(endpoint: str) -> str:
        return endpoint[:-1] if endpoint[-1] == '/' else endpoint

    ###############################################################################################################
    # Public methods
    ###############################################################################################################

    def get_api_endpoint_of(self, api_name: str) -> str:
        """
        Determines endpoints of the API names.
        Loads and caches the current API information if necessary.

        :param api_name: Name of the HIRO API
        :return: Full url of endpoint for this API
        """

        if self.custom_endpoints:
            endpoint = self.custom_endpoints.get(api_name)
            if endpoint:
                return self._remove_slash(self._root_url + endpoint)

        if not self._version_info:
            self._version_info = self.get_version()

        api_entry: dict = self._version_info.get(api_name)

        if not api_entry:
            raise ValueError("No API named '{}' found.".format(api_name))

        return self._remove_slash(self._root_url + api_entry.get('endpoint'))

    def get_websocket_config(self, api_name: str) -> Tuple[
        str,
        str,
        Optional[str],
        Optional[int],
        Optional[dict]
    ]:
        """
        Determines endpoints for websockets of the API names.
        Loads and caches the current API information if necessary.
        If proxies have been given, the key of the proxy picked needs to be "ws" or "wss" respectively.

        :param api_name: Name of the HIRO API for websockets
        :return: Tuple of full url of websocket for this API, its protocol, its proxy_host, its proxy port and proxy
                 auth (if any).
        """

        def _get_proxy_info(_url: str) -> Tuple[Optional[str], Optional[int], Optional[dict]]:
            proxies = self._get_proxies()
            if not proxies:
                return None, None, None

            url_parts = urllib.parse.urlparse(_url)
            proxy_url = proxies.get(url_parts.scheme)

            if not proxy_url:
                return None, None, None

            proxy_url_parts = urllib.parse.urlparse(proxy_url)

            proxy_auth: dict = {
                proxy_url_parts.username: proxy_url_parts.password
            } if proxy_url_parts.username else None

            return proxy_url_parts.hostname, proxy_url_parts.port, proxy_auth

        def _construct_result(_endpoint: str, _protocol: str) -> Tuple[
            str,
            str,
            Optional[str],
            Optional[str],
            Optional[dict]
        ]:
            _url: str = self._root_url.lower().replace('https://', 'wss://').replace('http://', 'ws://')
            _proxy, _proxy_port, _proxy_auth = _get_proxy_info(_url)
            return self._remove_slash(_url + _endpoint), _protocol, _proxy, _proxy_port, _proxy_auth

        if self.custom_endpoints:
            value = self.custom_endpoints.get(api_name)
            if isinstance(value, tuple):
                endpoint, protocol = value
                if endpoint:
                    return _construct_result(endpoint, protocol)

        if not self._version_info:
            self._version_info = self.get_version()

        api_entry: dict = self._version_info.get(api_name)

        if not api_entry:
            raise ValueError("No WS API named '{}' found.".format(api_name))

        return _construct_result(api_entry.get('endpoint'), api_entry.get('protocol'))

    ###############################################################################################################
    # REST API operations
    ###############################################################################################################

    def get_version(self) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/api/version'`

        :return: The result payload
        """
        url = self._root_url + '/api/version'
        return self.get(url)

    ###############################################################################################################
    # Token handling
    ###############################################################################################################

    @property
    def token(self) -> str:
        """
        Return the current token.
        :return: The current token
        """
        raise RuntimeError('Cannot use property of this abstract class.')

    @abstractmethod
    def refresh_token(self) -> None:
        """
        Refresh the current token.
        """
        raise RuntimeError('Cannot use method of this abstract class.')

    @abstractmethod
    def refresh_time(self) -> Optional[int]:
        """
        Calculate the time after which the token should be refreshed in milliseconds.

        :return: The timestamp in ms after which the token shall be refreshed or None if the token cannot be refreshed
                 on its own.
        """
        raise RuntimeError('Cannot use method of this abstract class.')


class FixedTokenApiHandler(AbstractTokenApiHandler):
    """
    TokenApiHandler for a fixed token.
    """

    _token: str

    def __init__(self,
                 root_url: str,
                 token: str,
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = 600,
                 client_name: str = None,
                 custom_endpoints: dict = None):
        """
        Constructor

        :param root_url: Root url for HIRO, like https://core.arago.co.
        :param token: The fixed token to use.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
                            is given with another value for "User-Agent". Default is "python-hiro-client".
        :param custom_endpoints: Optional map of [name:endpoint_path] that overrides or adds to the endpoints taken from
               /api/version.
        """
        super().__init__(
            root_url=root_url,
            raise_exceptions=raise_exceptions,
            proxies=proxies,
            timeout=timeout,
            headers=headers,
            client_name=client_name,
            custom_endpoints=custom_endpoints
        )

        self._token = token

    @property
    def token(self) -> str:
        return self._token

    def refresh_token(self) -> None:
        raise FixedTokenError('Token is invalid and cannot be changed because it has been given externally.')

    def refresh_time(self) -> Optional[int]:
        """

        :return: Always none
        """
        return None


class EnvironmentTokenApiHandler(AbstractTokenApiHandler):
    """
    TokenApiHandler for a fixed token given as an environment variable.
    """

    _env_var: str

    def __init__(self,
                 root_url: str,
                 env_var: str = 'HIRO_TOKEN',
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = 600,
                 client_name: str = None,
                 custom_endpoints: dict = None):
        """
        Constructor

        :param root_url: Root url for HIRO, like https://core.arago.co.
        :param env_var: Name of the environment variable to read for the token. Default is *HIRO_TOKEN*.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
                            is given with another value for "User-Agent". Default is "python-hiro-client".
        :param custom_endpoints: Optional map of [name:endpoint_path] that overrides or adds to the endpoints taken from
               /api/version.
        """
        super().__init__(
            root_url=root_url,
            raise_exceptions=raise_exceptions,
            proxies=proxies,
            headers=headers,
            timeout=timeout,
            client_name=client_name,
            custom_endpoints=custom_endpoints
        )

        self._env_var = env_var

    @property
    def token(self) -> str:
        return os.environ[self._env_var]

    def refresh_token(self) -> None:
        raise FixedTokenError(
            "Token is invalid and cannot be changed because it has been given as environment variable '{}'"
            " externally.".format(self._env_var))

    def refresh_time(self) -> Optional[int]:
        """

        :return: Always none
        """
        return None


class TokenInfo:
    """
    This class stores token information from the auth api.
    """

    token: str = None
    """ The token string """
    expires_at = -1
    """ Token expiration in ms since epoch"""
    refresh_token: str = None
    """ The refresh token to use - if any."""
    last_update = 0
    """ Timestamp of when the token has been fetched in ms."""
    refresh_offset = 5000
    """ Milliseconds of offset for token expiry """

    def __init__(self, token: str = None, refresh_token: str = None, expires_at: int = -1, refresh_offset: int = 5000):
        """
        Constructor

        :param token: The token string
        :param refresh_token: A refresh token
        :param expires_at: Token expiration in ms since epoch
        :param refresh_offset: Offset in milliseconds that will be subtracted from the expiry time so a token will be
                               refreshed in time. Default is 5 seconds.
        """
        self.token = token
        self.expires_at = expires_at
        self.refresh_token = refresh_token
        self.last_update = self.get_epoch_millis() if token else 0
        self.refresh_offset = refresh_offset

    @staticmethod
    def get_epoch_millis() -> int:
        """
        Get timestamp
        :return: Current epoch in milliseconds
        """
        return int(time.time_ns() / 1000000)

    def parse_token_result(self, res: dict, what: str) -> None:
        """
        Parse the result payload and extract token information.

        :param res: The result payload from the backend.
        :param what: What token command has been issued (for error messages).
        :raises TokenUnauthorizedError: When the token request returned error 401. This usually means, that this token
                has expired.
        :raises AuthenticationTokenError: When the token request returned any other error.
        """
        if 'error' in res:
            message: str = '{}: {}'.format(what, res['error'].get('message'))
            code: int = int(res['error'].get('code'))

            if code == 401:
                raise TokenUnauthorizedError(message, code)
            else:
                raise AuthenticationTokenError(message, code)

        self.last_update = self.get_epoch_millis()

        self.token = res.get('_TOKEN')

        expires_at = res.get('expires-at')
        if expires_at:
            self.expires_at = int(expires_at)
        else:
            expires_in = res.get('expires_in')
            if expires_in:
                self.expires_at = self.last_update + int(expires_in) * 1000

        refresh_token = res.get('refresh_token')
        if refresh_token:
            self.refresh_token = refresh_token

    def expired(self) -> bool:
        """
        Check token expiration

        :return: True when the token has been expired *(expires_at - refresh_offset) <= get_epoch_mills()*
        """
        return self.refresh_time() <= self.get_epoch_millis()

    def fresh(self, span: int = 30000) -> bool:
        """
        Check, whether the last token fetched is younger than span ms.

        :param span: Timespan in ms in which a token is considered fresh. Default is 30 sec (30000ms).
        :return: True when the last update was less than span ms.
        """

        return (self.get_epoch_millis() - self.last_update) < span

    def refresh_time(self) -> Optional[int]:
        """
        Calculate the time after which the token should be refreshed in milliseconds.

        :return: expires_at - refresh_offset (in ms) or None if refresh is not possible.
        """
        return self.expires_at - self.refresh_offset if self.expires_at > 0 else None


class PasswordAuthTokenApiHandler(AbstractTokenApiHandler):
    """
    API Tokens will be fetched using this class. It does not handle any automatic token fetching, refresh or token
    expiry. This has to be checked and triggered by the *caller*.

    The methods of this class are thread-safe so it can be shared between several HIRO objects.

    It is built this way to avoid endless calling loops when resolving tokens.
    """

    _token_info: TokenInfo = None
    """Contains all token information"""

    _lock: threading.RLock
    """Reentrant mutex for thread safety"""

    _username: str
    _password: str
    _client_id: str
    _client_secret: str

    _secure_logging: bool = True

    def __init__(self,
                 root_url: str,
                 username: str,
                 password: str,
                 client_id: str,
                 client_secret: str,
                 secure_logging: bool = True,
                 raise_exceptions: bool = True,
                 proxies: dict = None,
                 headers: dict = None,
                 timeout: int = 600,
                 client_name: str = None,
                 custom_endpoints: dict = None):
        """
        Constructor

        :param root_url: Root url for HIRO, like https://core.arago.co.
        :param username: Username for authentication
        :param password: Password for authentication
        :param client_id: OAuth client_id for authentication
        :param client_secret: OAuth client_secret for authentication
        :param secure_logging: If this is enabled, payloads that might contain sensitive information are not logged.
        :param raise_exceptions: Raise exceptions on HTTP status codes that denote an error. Default is True.
        :param proxies: Proxy configuration for *requests*. Default is None.
        :param headers: Optional custom HTTP headers. Will override the internal headers. Default is None.
        :param timeout: Optional timeout for requests. Default is 600 (10 min).
        :param client_name: Optional name for the client. Will also be part of the "User-Agent" header unless *headers*
                            is given with another value for "User-Agent". Default is "python-hiro-client".
        :param custom_endpoints: Optional map of [name:endpoint_path] that overrides or adds to the endpoints taken from
               /api/version.
        """
        super().__init__(
            root_url=root_url,
            raise_exceptions=raise_exceptions,
            proxies=proxies,
            headers=headers,
            timeout=timeout,
            client_name=client_name,
            custom_endpoints=custom_endpoints
        )

        self._username = username
        self._password = password
        self._client_id = client_id
        self._client_secret = client_secret

        self._secure_logging = secure_logging

        self._token_info = TokenInfo()
        self._lock = threading.RLock()

    @property
    def endpoint(self):
        return self.get_api_endpoint_of('auth')

    @property
    def token(self) -> str:
        with self._lock:
            if not self._token_info.token:
                self.get_token()
            elif self._token_info.expired():
                self.refresh_token()

            return self._token_info.token

    def _log_communication(self, res: requests.Response, request_body: bool = True, response_body: bool = True) -> None:
        """
        Logging under a secure aspect. Hides sensitive information unless *self._secure_logging* is set to False.

        :param res: The response of a request. Also contains the request.
        :param request_body: Option to disable the logging of the request_body. If set to True, will only remain True
               internally when *self._secure_logging* is set to False.
        :param response_body: Option to disable the logging of the response_body.  If set to True, will only remain True
               internally when *self._secure_logging* is set to False or *res.status_code* != 200.
        """
        log_request_body = not self._secure_logging and request_body is True
        log_response_body = (res.status_code != 200 or not self._secure_logging) and response_body is True

        super()._log_communication(res, request_body=log_request_body, response_body=log_response_body)

    def get_token(self) -> None:
        """
        Construct a request to obtain a new token. API self._endpoint + '/app'

        :raises AuthenticationTokenError: When no auth_endpoint is set.
        """
        with self._lock:
            if not self.endpoint:
                raise AuthenticationTokenError(
                    'Token is invalid and endpoint (auth_endpoint) for obtaining is not set.')

            if not self._username or not self._password or not self._client_id or not self._client_secret:
                msg = ""
                if not self._username:
                    msg += "'username'"
                if not self._password:
                    msg += (", " if msg else "") + "'password'"
                if not self._client_id:
                    msg += (", " if msg else "") + "'client_id'"
                if not self._client_secret:
                    msg += (", " if msg else "") + "'client_secret'"
                raise AuthenticationTokenError(
                    "{} is missing required parameter(s) {}.".format(self.__class__.__name__, msg))

            url = self.endpoint + '/app'
            data = {
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "username": self._username,
                "password": self._password
            }

            res = self.post(url, data)
            self._token_info.parse_token_result(res, "{}.get_token".format(self.__class__.__name__))

    def refresh_token(self) -> None:
        """
        Construct a request to refresh an existing token. API self._endpoint + '/refresh'.
        Does not refresh tokens that are younger than 30 sec to avoid refresh storms on parallel connections.

        :raises AuthenticationTokenError: When no auth_endpoint is set.
        """
        with self._lock:
            if not self.endpoint:
                raise AuthenticationTokenError(
                    'Token is invalid and endpoint (auth_endpoint) for refresh is not set.')

            if self._token_info.fresh():
                return

            if not self._token_info.refresh_token:
                self.get_token()
                return

            url = self.endpoint + '/refresh'
            data = {
                "client_id": self._client_id,
                "client_secret": self._client_secret,
                "refresh_token": self._token_info.refresh_token
            }

            try:
                res = self.post(url, data)
                self._token_info.parse_token_result(res, "{}.refresh_token".format(self.__class__.__name__))
            except AuthenticationTokenError:
                self.get_token()

    def refresh_time(self) -> Optional[int]:
        """
        Calculate refresh time.

        :return: Timestamp after which the token becomes invalid.
        """
        return self._token_info.refresh_time()

    ###############################################################################################################
    # Response and token handling
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        This is a dummy method. No response checking here.

        :param res: The result payload
        """
        return

    def _handle_token(self) -> Optional[str]:
        """
        Just return None, therefore a header without Authorization
        will be created in *self._get_headers()*.

        Does *not* try to obtain or refresh a token.

        :return: *token* given.
        """
        return None


###################################################################################################################
# Root class for different API groups
###################################################################################################################

class AuthenticatedAPIHandler(AbstractAPI):
    """
    Python implementation for accessing a REST API with authentication.
    """

    _api_handler: AbstractTokenApiHandler
    _api_name: str

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 api_name: str):
        """
        Constructor

        :param api_name: Name of the API to use.
        :param api_handler: External API handler.
        """
        if not api_handler or not api_name:
            raise ValueError("Cannot authenticate against HIRO without *api_handler* and *api_name*.")

        super().__init__(root_url=api_handler._root_url,
                         raise_exceptions=api_handler._raise_exceptions,
                         proxies=api_handler._proxies,
                         headers=api_handler._headers,
                         timeout=api_handler._timeout,
                         client_name=api_handler._client_name)

        self._api_handler = api_handler
        self._api_name = api_name

    @property
    def endpoint(self) -> str:
        return self._api_handler.get_api_endpoint_of(self._api_name)

    ###############################################################################################################
    # Response and token handling
    ###############################################################################################################

    def _check_response(self, res: requests.Response) -> None:
        """
        Response checking. Tries to refresh the token on status_code 401, then raises RequestException to try
        again using backoff.

        :param res: The result payload
        :raises requests.exceptions.RequestException: When an error 401 occurred and the token has been refreshed.
        """
        if res.status_code == 401:
            self._api_handler.refresh_token()

            # Raise this exception to trigger retry with backoff
            raise requests.exceptions.RequestException

    def _handle_token(self) -> Optional[str]:
        """
        Try to return a valid token by obtaining or refreshing it.

        :return: A valid token.
        """
        return self._api_handler.token


###################################################################################################################
# Exceptions
###################################################################################################################

class AuthenticationTokenError(Exception):
    """
    Class for unrecoverable failures with access tokens.
    Contains a message and an optional message code. If the code is None, no code will be printed in __str__().
    """
    message: str
    code: int

    def __init__(self, message: str, code: int = None):
        self.message = message
        self.code = code

    def __str__(self) -> str:
        if self.code is None:
            return "{}: {}".format(self.__class__.__name__, self.message)
        else:
            return "{}: {} ({})".format(self.__class__.__name__, self.message, self.code)


class TokenUnauthorizedError(AuthenticationTokenError):
    """
    Child of *AuthenticationTokenErrors*. Used when tokens expire with error 401.
    """
    pass


class FixedTokenError(AuthenticationTokenError):
    """
    Child of *AuthenticationTokenErrors*. Used when tokens are fixed and cannot be refreshed.
    """
    pass



"""
Package which contains the classes to communicate with HIRO Graph.
"""
import site
from os import path

from hiro_graph_client.appclient import HiroApp
from hiro_graph_client.authclient import HiroAuth
from hiro_graph_client.batchclient import HiroGraphBatch, SessionData, AbstractIOCarrier, BasicFileIOCarrier, \
    SourceValueError, HiroResultCallback
from hiro_graph_client.client import HiroGraph
from hiro_graph_client.clientlib import AbstractTokenApiHandler, AuthenticationTokenError, FixedTokenError, \
    TokenUnauthorizedError, PasswordAuthTokenApiHandler, FixedTokenApiHandler, EnvironmentTokenApiHandler, \
    accept_all_certs

from hiro_graph_client.version import __version__

this_directory = path.abspath(path.dirname(__file__))

__all__ = [
    'HiroGraph', 'HiroAuth', 'HiroApp', 'HiroGraphBatch', 'SessionData',
    'HiroResultCallback', 'AbstractTokenApiHandler', 'PasswordAuthTokenApiHandler', 'FixedTokenApiHandler',
    'EnvironmentTokenApiHandler', 'AuthenticationTokenError', 'FixedTokenError', 'TokenUnauthorizedError',
    'AbstractIOCarrier', 'BasicFileIOCarrier', 'SourceValueError', '__version__', 'accept_all_certs'
]

site.addsitedir(this_directory)



#!/usr/bin/env python3

from typing import Any, Iterator
from urllib.parse import quote_plus

from hiro_graph_client.clientlib import AuthenticatedAPIHandler, AbstractTokenApiHandler


class HiroGraph(AuthenticatedAPIHandler):
    """
    Python implementation for accessing the HIRO HIRO REST API.
    See https://core.arago.co/help/specs/?url=definitions/graph.yaml
    """

    def __init__(self, api_handler: AbstractTokenApiHandler):
        """
        Constructor

        :param api_handler: External API handler.
        """
        super().__init__(api_name='graph',
                         api_handler=api_handler)

    ###############################################################################################################
    # REST API operations
    ###############################################################################################################

    def query(self,
              query: str,
              fields: str = None,
              limit=-1,
              offset=0,
              order: str = None,
              meta=False) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/query/vertices'`

        :param query: The actual query. e.g. ogit\\\\/_type: ogit\\\\/Question for vertices.
        :param fields: the comma separated list of fields to return
        :param limit: limit of entries to return
        :param offset: offset where to start returning entries
        :param order: order by a field asc|desc, e.g. ogit/name desc
        :param meta: List detailed metainformations in result payload
        :return: Result payload
        """
        url = self.endpoint + '/query/vertices'
        data = {"query": str(query),
                "limit": limit,
                "fields": (quote_plus(fields.replace(" ", ""), safe="/,") if fields else ""),
                "count": False,
                "listMeta": meta,
                "offset": offset}
        if order is not None:
            data['order'] = order
        return self.post(url, data)

    def query_gremlin(self,
                      query: str,
                      root: str,
                      fields: str = None,
                      include_deleted: bool = False,
                      meta=False) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/query/gremlin'`

        :param query: The actual query. e.g. outE().inV() for gremlin.
        :param root: ogit/_id of the root node where the gremlin query starts.
        :param fields: the comma separated list of fields to return
        :param include_deleted: Include deleted values.
        :param meta: List detailed metainformations in result payload
        :return: Result payload
        """
        url = self.endpoint + '/query/gremlin'
        data = {"query": str(query),
                "root": root,
                "fields": (quote_plus(fields.replace(" ", ""), safe="/,") if fields else ""),
                "includeDeleted": include_deleted,
                "listMeta": meta}
        return self.post(url, data)

    def create_node(self, data: dict, obj_type: str, return_id=False) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/new/{id}'`

        :param data: Payload for the new node/vertex
        :param obj_type: ogit/_type of the new node/vertex
        :param return_id: Return only the ogit/_id. Default is False to return everything.
        :return: The result payload
        """
        url = self.endpoint + '/new/' + quote_plus(obj_type)
        res = self.post(url, data)
        return res['ogit/_id'] if return_id and 'error' not in res else res

    def update_node(self, node_id: str, data: dict) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/{id}'`

        :param data: Payload for the node/vertex
        :param node_id: ogit/_id of the node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id)
        return self.post(url, data)

    def delete_node(self, node_id: str) -> dict:
        """
        HIRO REST query API: `DELETE self._endpoint + '/{id}'`

        :param node_id: ogit/_id of the node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id)
        return self.delete(url)

    def connect_nodes(self, from_node_id: str, verb: str, to_node_id: str) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/connect/{verb}'`

        :param from_node_id: ogit/_id of the source node/vertex
        :param verb: verb for the connection
        :param to_node_id: ogit/_id of the target node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/connect/' + quote_plus(verb)
        data = {"out": from_node_id, "in": to_node_id}
        return self.post(url, data)

    def disconnect_nodes(self, from_node_id: str, verb: str, to_node_id: str) -> dict:
        """
        HIRO REST query API: `DELETE self._endpoint + '/{id}'`

        :param from_node_id: ogit/_id of the source node/vertex
        :param verb: verb for the connection
        :param to_node_id: ogit/_id of the target node/vertex
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(
            from_node_id
        ) + "$$" + quote_plus(
            verb
        ) + "$$" + quote_plus(
            to_node_id
        )
        return self.delete(url)

    def get_node(self, node_id: str, fields: str = None, meta: bool = None) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}'`

        :param node_id: ogit/_id of the node/vertex or edge
        :param fields: Filter for fields
        :param meta: List detailed metainformations in result payload
        :return: The result payload
        """
        query = {
            "fields": fields.replace(" ", "") if fields else None,
            "listMeta": "true" if meta else None
        }

        url = self.endpoint + '/' + quote_plus(node_id) + self._get_query_part(query)
        return self.get(url)

    def get_nodes(self, node_ids: list, fields: str = None, meta: bool = None) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}'`

        :param node_ids: list of ogit/_ids of the node/vertexes or edges
        :param fields: Filter for fields
        :param meta: List detailed metainformations in result payload
        :return: The result payload
        """
        query = {
            "query": ",".join(node_ids),
            "fields": fields.replace(" ", "") if fields else None,
            "listMeta": "true" if meta else None
        }

        url = self.endpoint + '/query/ids' + self._get_query_part(query)
        return self.get(url)

    def get_node_by_xid(self, node_id: str, fields: str = None, meta: bool = None) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/xid/{xid}'`

        :param node_id: ogit/_xid of the node/vertex or edge
        :param fields: Filter for fields
        :param meta: List detailed metainformations in result payload
        :return: The result payload
        """
        query = {
            "fields": fields.replace(" ", "") if fields else None,
            "listMeta": "true" if meta else None
        }

        url = self.endpoint + '/xid/' + quote_plus(node_id) + self._get_query_part(query)
        return self.get(url)

    def get_timeseries(self, node_id: str, starttime: str = None, endtime: str = None) -> dict:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}/values'`

        :param node_id: ogit/_id of the node containing timeseries
        :param starttime: ms since epoch.
        :param endtime: ms since epoch.
        :return: The result payload
        """
        query = {
            "from": starttime,
            "to": endtime
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/values' + self._get_query_part(query)
        res = self.get(url)
        if 'error' in res:
            return res
        timeseries = res['items']
        timeseries.sort(key=lambda x: x['timestamp'])
        return timeseries

    def post_timeseries(self, node_id: str, items: list) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/{id}/values'`

        :param node_id: ogit/_id of the node containing timeseries
        :param items: list of timeseries values [{timestamp: (ms since epoch), value: ...},...]
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id) + '/values?synchronous=true'
        data = {"items": items}
        return self.post(url, data)

    def get_attachment(self,
                       node_id: str,
                       content_id: str = None,
                       include_deleted: bool = False) -> Iterator[bytes]:
        """
        HIRO REST query API: `GET self._endpoint + '/{id}/content'`

        :param node_id: Id of the attachment node
        :param content_id: Id of the content within the attachment node. Default is None.
        :param include_deleted: Whether to be able to access deleted content: Default is False
        :return: Returns generator over byte chunks from the response body payload.
        """
        query = {
            "contentId": content_id,
            "includeDeleted": "true" if include_deleted else None
        }

        url = self.endpoint + '/' + quote_plus(node_id) + '/content' + self._get_query_part(query)
        yield from self.get_binary(url)

    def post_attachment(self,
                        node_id: str,
                        data: Any,
                        content_type: str = None) -> dict:
        """
        HIRO REST query API: `POST self._endpoint + '/{id}/content'`

        :param node_id: Id of the attachment node
        :param data: Data to upload in binary form. Can also be an IO object for streaming.
        :param content_type: Content-Type for *data*. Defaults to 'application/octet-stream' if left unset.
        :return: The result payload
        """
        url = self.endpoint + '/' + quote_plus(node_id) + '/content'
        return self.post_binary(url, data, content_type=content_type)



import json
import logging
import threading
from datetime import datetime
from typing import List, Dict

from apscheduler.schedulers.background import BackgroundScheduler
from websocket import WebSocketApp, WebSocketException

from hiro_graph_client.clientlib import AbstractTokenApiHandler
from hiro_graph_client.websocketlib import AbstractAuthenticatedWebSocketHandler, ErrorMessage, ReaderStatus

logger = logging.getLogger(__name__)
""" The logger for this module """


class EventMessage:
    """
    The structure of an incoming events message
    """
    id: str
    timestamp: int
    nanotime: int
    body: dict
    type: str
    metadata: dict

    def __init__(self,
                 event_id: str,
                 event_timestamp: int,
                 event_body: dict,
                 event_type: str,
                 event_metadata: dict,
                 event_nanotime: int):
        """
        Constructor

        :param event_id: ID
        :param event_timestamp: Timestamp in milliseconds
        :param event_body: Body dict
        :param event_type: Type of event. CREATE, UPDATE or DELETE.
        :param event_metadata: Additional metadata.
        :param event_nanotime: Nanotime for the event
        """
        self.id = event_id
        self.timestamp = event_timestamp
        self.body = event_body
        self.type = event_type.upper()
        self.metadata = event_metadata
        self.nanotime = event_nanotime

    @classmethod
    def parse(cls, message: str):
        """
        :param message: The message received from the websocket. Will be decoded here.
        :return: The EventMessage or None if this is not an EventMessage (type or id are missing).
        """
        json_message: dict = json.loads(message)
        if not isinstance(json_message, dict):
            return None

        event_type = json_message.get('type')
        event_id = json_message.get('id')
        if not event_type or not event_id:
            return None

        return cls(event_id,
                   json_message.get('timestamp'),
                   json_message.get('body'),
                   event_type,
                   json_message.get('metadata'),
                   json_message.get('nanotime'))

    def __str__(self):
        return json.dumps(vars(self))


class EventsFilter:
    """
    The event filter structure
    """
    id: str
    type: str
    content: str

    def __init__(self, filter_id: str, filter_content: str, filter_type: str = None):
        """
        Constructor

        :param filter_id: Unique name/id of the filter
        :param filter_content: jfilter specification for the filter.
        :param filter_type: Type of filter. Only 'jfilter' (the default when this is None) is possible here atm.
        """
        self.id = filter_id
        self.content = filter_content
        self.type = filter_type or 'jfilter'

    def __str__(self):
        return json.dumps(vars(self))

    def to_dict(self) -> dict:
        return {
            "filter-id": self.id,
            "filter-type": self.type,
            "filter-content": self.content
        }


class AbstractEventsWebSocketHandler(AbstractAuthenticatedWebSocketHandler):
    """
    A handler for issue events
    """
    _events_filter_messages: Dict[str, EventsFilter] = {}
    _scopes: List[str] = []

    _initial_messages_lock: threading.RLock
    """Lock for _events_filter_messages and _scopes """

    _token_scheduler: BackgroundScheduler

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 events_filters: List[EventsFilter],
                 scopes: List[str] = None,
                 query_params: Dict[str, str] = None):
        """
        Constructor

        :param api_handler: The TokenApiHandler for this WebSocket.
        :param events_filters: Filters for the events. These have to be set or the flood of incoming events will be too
                               big.
        :param scopes: List of ids of non-default scopes to subscribe. These are ogit/_id of the "ogit/DataScope"s
                       (i.e. the scope of your instance) you want to subscribe to. Default is None, which means:
                       Use default scope.
        :param query_params: URL Query parameters for this specific websocket. Use Dict[str,str] only here,
                             i.e set {"allscopes": "false"} instead of {"allscopes": False}. The default here is to set
                             {'allscopes': 'false'}.
        """
        super().__init__(api_handler,
                         'events-ws',
                         query_params=query_params if query_params is not None else {'allscopes': 'false'})

        self._initial_messages_lock = threading.RLock()

        self._token_scheduler = BackgroundScheduler()

        if logging.root.level == logging.INFO:
            logging.getLogger('apscheduler').setLevel(logging.WARNING)

        if not events_filters:
            raise ValueError('Parameter events_filters= cannot be empty. It needs at least one EventsFilter.')

        for events_filter in events_filters:
            self._events_filter_messages[events_filter.id] = events_filter

        self._scopes = scopes or []

    ###############################################################################################################
    # Websocket Events
    ###############################################################################################################

    def on_open(self, ws: WebSocketApp):
        """
        Register the filters when websocket opens. If this fails, the websocket gets closed again.

        :param ws: The WebSocketApp
        :raise WebSocketFilterException: When setting the filters failed.
        """
        try:
            if not self._token_scheduler.running:
                self._set_next_token_refresh()
                self._token_scheduler.start()

            with self._initial_messages_lock:
                for scope in self._scopes:
                    scope_message = self._get_subscribe_scope_message(scope)
                    self.send(scope_message)

                for events_filter in self._events_filter_messages.values():
                    filter_message = self._get_events_register_message(events_filter)
                    self.send(filter_message)

        except Exception as err:
            raise WebSocketFilterException('Setting events filter failed') from err

    def on_close(self, ws: WebSocketApp, code: int = None, reason: str = None):
        """
        Cancel the self._token_refresh_thread. Registered filters remain as they are.

        :param ws: The WebSocketApp
        :param code:
        :param reason:
        """
        if self._token_scheduler.running:
            self._token_scheduler.shutdown()

    def on_message(self, ws: WebSocketApp, message: str):
        """
        Create an EventMessage from the incoming message and hand it over to *self.on_event*.

        :param ws: The WebSocketApp
        :param message: The raw message as string
        """

        event_message = EventMessage.parse(message)
        if event_message:
            if event_message.type not in ['CREATE', 'UPDATE', 'DELETE']:
                logger.error("Unknown event message of type '%s'", event_message.type)
            else:
                self.on_event(event_message)
        else:
            error_message = ErrorMessage.parse(message)
            if error_message:
                logger.error("Received error: %s", str(error_message))
                if self._reader_status == ReaderStatus.RUNNING_PRELIMINARY:
                    self._reader_status = ReaderStatus.FAILED
            else:
                logger.error("Invalid message: %s", message)

    def on_error(self, ws: WebSocketApp, error: Exception):
        """
        Does nothing here.

        :param ws: The WebSocketApp
        :param error: Exception
        """
        pass

    ###############################################################################################################
    # Public API Reader thread
    ###############################################################################################################

    def on_event(self, message: EventMessage) -> None:
        """
        Catches all event messages. Distributes them between *self.on_create*, *self.on_update* or *self.on_delete*
        by default.
        Overwrite this if you want a catch-all for all event messages.

        :param message: The incoming EventMessage
        """
        if message.type == 'CREATE':
            self.on_create(message)
        elif message.type == 'UPDATE':
            self.on_update(message)
        elif message.type == 'DELETE':
            self.on_delete(message)

    def on_create(self, message: EventMessage) -> None:
        """
        Called by CREATE events. Skeleton function to be overwritten if needed.

        :param message: The incoming EventMessage
        """
        pass

    def on_update(self, message: EventMessage) -> None:
        """
        Called by UPDATE events. Skeleton function to be overwritten if needed.

        :param message: The incoming EventMessage
        """
        pass

    def on_delete(self, message: EventMessage) -> None:
        """
        Called by DELETE events. Skeleton function to be overwritten if needed.

        :param message: The incoming EventMessage
        """
        pass

    ###################################################################################################################
    # Filter handling
    ###################################################################################################################

    @staticmethod
    def _get_events_register_message(events_filter: EventsFilter) -> str:
        message: dict = {
            "type": "register",
            "args": events_filter.to_dict()
        }

        return json.dumps(message)

    @staticmethod
    def _get_subscribe_scope_message(scope_id: str) -> str:
        message: dict = {
            "type": "subscribe",
            "id": scope_id
        }

        return json.dumps(message)

    def add_events_filter(self, events_filter: EventsFilter) -> None:
        message: str = self._get_events_register_message(events_filter)
        self.send(message)
        with self._initial_messages_lock:
            self._events_filter_messages[events_filter.id] = events_filter

    def remove_events_filter(self, events_filter_id: str) -> None:
        message: dict = {
            "type": "unregister",
            "args": {
                "filter-id": events_filter_id
            }
        }

        self.send(json.dumps(message))
        with self._initial_messages_lock:
            del self._events_filter_messages[events_filter_id]

    def clear_events_filters(self) -> None:
        message: dict = {
            "type": "clear",
            "args": {}
        }

        self.send(json.dumps(message))
        with self._initial_messages_lock:
            self._events_filter_messages = {}

    def subscribe_scope(self, scope_id: str):
        message = self._get_subscribe_scope_message(scope_id)
        self.send(json.dumps(message))
        with self._initial_messages_lock:
            self._scopes.append(scope_id)

    def remove_scope(self, scope_id: str):
        """
        This only removes the scope from the internal list since there is no 'unsubscribe'. You need to
        restart the websocket for this change to take effect.
        """
        with self._initial_messages_lock:
            self._scopes.remove(scope_id)

    ###################################################################################################################
    # Token refresh thread
    ###################################################################################################################

    def _set_next_token_refresh(self):
        if self._api_handler.refresh_time() is not None:
            # make seconds
            timestamp = self._api_handler.refresh_time() / 1000

            self._token_scheduler.add_job(
                func=lambda: self._token_refresh_thread(),
                trigger='date',
                run_date=datetime.fromtimestamp(timestamp),
                id='token_refresh_thread',
                replace_existing=True)

    def _token_refresh_thread(self):
        logger.debug("Updating token for session")

        message: dict = {
            "type": "token",
            "args": {
                "_TOKEN": self._api_handler.token
            }
        }

        self.send(json.dumps(message))

        self._set_next_token_refresh()

    ###################################################################################################################
    # Sending messages
    ###################################################################################################################
    #
    # This is not needed, since events are only received, not sent.
    #
    # def send_events_message(self, events_type: str, headers: dict, body: dict) -> str:
    #     uuid: str = uuid4().hex
    #
    #     message: dict = {
    #         "id": uuid,
    #         "type": events_type,
    #         "headers": headers,
    #         "body": body
    #     }
    #
    #     self.send(json.dumps(message))
    #
    #     return uuid


###################################################################################################################
# Exceptions
###################################################################################################################

class WebSocketFilterException(WebSocketException):
    """
    On errors with setting or parsing filter information.
    """
    pass



import concurrent.futures
import json
import logging
import random
import ssl
import threading
import time
from abc import abstractmethod
from enum import Enum
from typing import Optional, Tuple, List, Dict
from urllib.parse import urlencode

from websocket import WebSocketApp, ABNF, WebSocketException, setdefaulttimeout, WebSocketConnectionClosedException, \
    STATUS_NORMAL, STATUS_UNEXPECTED_CONDITION

from hiro_graph_client.clientlib import AbstractTokenApiHandler

logger = logging.getLogger(__name__)
""" The logger for this module """


class ErrorMessage:
    """
    The structure of an incoming error message
    """
    code: int
    message: str

    def __init__(self,
                 code,
                 message):
        """
        Constructor

        :param code: Numerical error code of the error
        :param message: Error message
        """
        self.code = int(code)
        self.message = str(message)

    @classmethod
    def parse(cls, message: str):
        """
        :param message: The message received from the websocket. Will be decoded here.
        :return: The new error message or None if this is not an error message.
        """
        json_message: dict = json.loads(message)
        error_message = json_message.get('error')
        if isinstance(error_message, dict):
            return cls(error_message.get('code'),
                       error_message.get('message'))
        else:
            return None

    def __str__(self):
        return json.dumps(vars(self))

    def to_dict(self) -> dict:
        return {
            "error": {
                "code": self.code,
                "message": self.message
            }
        }


class ReaderStatus(str, Enum):
    """
    The states the reader thread can be in.
    """
    NONE = 'Not started',
    STARTING = 'Starting',
    RUNNING_PRELIMINARY = 'Running preliminary (status of token unknown)',
    RUNNING = 'Running'
    RESTARTING = 'Restarting',
    DONE = 'Finished normally',
    FAILED = 'Finished because of error'


class AbstractAuthenticatedWebSocketHandler:
    """
    The basic class for all WebSockets.
    """
    _api_handler: AbstractTokenApiHandler
    _proxy_hostname: str
    _proxy_port: str
    _proxy_auth: dict

    _reconnect_delay: int

    _protocol: str
    _url: str

    _timeout: int
    _auto_reconnect: bool

    _ws: Optional[WebSocketApp]
    _ws_lock: threading.RLock

    _reader_executor: concurrent.futures.ThreadPoolExecutor
    """
    Executor for the reader thread *self._run()*.
    """
    _reader_future: Optional[concurrent.futures.Future]
    """
    Carries the result of the reader thread *self._run()* via the *self._reader_executor*.
    """
    _reader_status: ReaderStatus
    """
    Tracks the status of the internal reader thread.  
    """
    _reader_guard: threading.Condition
    """
    Meant to protect the startup sequence. *self.start()* will return only after the startup sequence has been finished.
    When start is handled, it gets notified in *self.on_open()* on success or *self.on_error()* on error.
    Se also *self._reader_starting*.
    """
    _inner_exception: Optional[Exception]
    """
    The WebSocketApp catches all exceptions that are thrown in the *on_[...]* methods, so we have to store the
    exceptions here.
    """

    _restart_lock: threading.RLock
    """ Only one restart attempt at one time """

    _remote_exit_codes: List[int] = []
    """ A list of remote exit codes that will cause the websocket to not reconnect."""

    _wait_until_finish: threading.Condition
    """ This condition gets notified when the reader thread returned and has been joined. """

    MAX_RETRIES = 3

    def __init__(self,
                 api_handler: AbstractTokenApiHandler,
                 api_name: str,
                 query_params: Dict[str, str] = None,
                 timeout: int = 5,
                 auto_reconnect: bool = True,
                 remote_exit_codes: List[int] = None):
        """
        Create the websocket

        :param api_handler: Required: The api handler for authentication.
        :param api_name: The name of the ws api.
        :param query_params: URL Query parameters for this specific websocket. Use Dict[str,str] only here,
                             i.e set {"allscopes": "false"} instead of {"allscopes": False}.
        :param timeout: The timeout for websocket messages. Default is 5sec.
        :param auto_reconnect: Try to create a new websocket automatically when *self.send()* fails. If this is set
                               to False, a WebSocketException will be raised instead. The default is True.
        :param remote_exit_codes: A list of close codes that can come in from the remote side. If one of the codes in
                                  this list matches the remote close code, reader thread will exit instead of
                                  trying to reconnect. The default is None -> reconnect always.
                                  See https://datatracker.ietf.org/doc/html/rfc6455#section-7.4.1
        """
        if not api_handler:
            raise ValueError('Parameter api_handler= cannot be empty.')

        if not api_name:
            raise ValueError('Parameter api_name= cannot be empty.')

        self._url, self._protocol, self._proxy_hostname, self._proxy_port, self._proxy_auth = \
            api_handler.get_websocket_config(api_name)

        if query_params:
            self._url += '?' + urlencode(query_params, safe="/,")

        self._api_handler = api_handler

        self._timeout = timeout
        self._auto_reconnect = auto_reconnect

        self._reader_guard = threading.Condition()
        self._reader_executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self._reader_status = ReaderStatus.NONE
        self._reader_future = None
        self._ws = None
        self._ws_lock = threading.RLock()
        self._inner_exception = None

        self._restart_lock = threading.RLock()

        if remote_exit_codes:
            self._remote_exit_codes = remote_exit_codes

        setdefaulttimeout(timeout)

        self._wait_until_finish = threading.Condition()

        random.seed(time.time_ns())

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.stop(self._timeout)

    def _set_error(self, error: Exception) -> None:
        """
        Log the Exception, store it and set the status to FAILED.

        :param error: The Exception to store.
        """
        self._reader_status = ReaderStatus.FAILED
        self._inner_exception = error
        logger.error("Reader thread encountered error: %s", str(error))

    def _check_message(self, ws: WebSocketApp, message: str) -> None:
        """
        Look for error 401. Try to reconnect with a new token when this is encountered.
        Set status to *ReaderStatus.RESTART* when a token is no longer valid on error code 401.
        Set status to *ReaderStatus.FAILED* when error 401 is received and the token was never valid.

        :param ws: WebSocketApp
        :param message: Incoming message as string
        """
        with self._reader_guard:
            try:
                error_message = ErrorMessage.parse(message)
                if error_message:
                    if error_message.code == 401:
                        if self._reader_status == ReaderStatus.RUNNING_PRELIMINARY:
                            raise WebSocketException(
                                "Received error message while token was never valid: " + str(error_message))
                        else:
                            self._api_handler.refresh_token()

                            # If we get here, the token has been refreshed successfully.
                            self._reconnect_delay = 0
                            self._reader_status = ReaderStatus.RESTARTING

                            logger.info("Refreshing token because of error: %s", str(error_message))
                            self._close(reason=f"{self._api_handler.user_agent} token refresh.")
                            return
                else:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Received message: %s", message)
                    else:
                        logger.info("Received message of length %d", len(message))

                self.on_message(ws, message)

                # If we get here, the token is valid
                if self._reader_status not in [ReaderStatus.RUNNING, ReaderStatus.FAILED]:
                    self._reader_status = ReaderStatus.RUNNING
                    self._reconnect_delay = 0

            except Exception as err:
                self._set_error(err)
                self._close(status=STATUS_UNEXPECTED_CONDITION,
                            reason="Exception while handling incoming text message.")

    def _check_open(self, ws: WebSocketApp) -> None:
        """
        First signal, that the connection has been opened successfully. Then
        call *self.on_open* and set *self._reader_status* to CHECKING_TOKEN if opening succeeded.

        :param ws: WebSocketApp
        """
        logger.debug("Connection to %s open.", self._url)

        with self._reader_guard:
            try:
                if self._reader_status == ReaderStatus.STARTING:
                    self._reader_guard.notify()
                self._reader_status = ReaderStatus.RUNNING_PRELIMINARY

                self.on_open(ws)
            except Exception as err:
                self._set_error(err)
                self._close(status=STATUS_UNEXPECTED_CONDITION,
                            reason="Exception while handling open message.")

    def _check_close(self, ws: WebSocketApp, code: int, reason: str):
        """
        Call *self.on_close*. When *code* and *reason* are None, the stop has been issued locally and not by the
        remote side.

        :param ws: WebSocketApp
        :param code: Code of stop message
        :param reason: Reason str of stop message
        """
        with self._reader_guard:
            try:
                if code or reason:
                    if code in self._remote_exit_codes or self._reader_status in [ReaderStatus.DONE,
                                                                                  ReaderStatus.FAILED]:
                        logger.debug("Received close from remote: %s %s. Closing...", code, reason)
                    else:
                        logger.debug("Received close from remote: %s %s. Restarting...", code, reason)
                        self._reader_status = ReaderStatus.RESTARTING
                else:
                    logger.debug("Received local close. Exiting...")

                self.on_close(ws, code, reason)
            except Exception as err:
                self._set_error(err)

    def _check_error(self, ws: WebSocketApp, error: Exception) -> None:
        """
        Just log the error and propagate it to *self.on_error*.

        :param ws: WebSocketApp
        :param error: Exception
        """
        # logger.error("Received error: %s", str(error))
        # logger.exception(error)

        with self._reader_guard:
            try:
                if self._reader_status == ReaderStatus.STARTING:
                    self._reader_guard.notify()

                self._set_error(error)
                self.on_error(ws, error)
            except Exception as err:
                self._set_error(err)

    def _run(self) -> Tuple[ReaderStatus, Exception]:
        """
        The _run loop of the reader thread. Exits only when *self._reader_status* is ReaderStatus.DONE or
        ReaderStatus.FAILED.

        :return: Tuple of (ReaderStatus - either DONE or ERROR, the Exception on ERROR or None)
        """
        with self._reader_guard:
            try:
                self._reconnect_delay = 0

                while self._reader_status not in [ReaderStatus.DONE, ReaderStatus.FAILED]:
                    self._reader_status = ReaderStatus.STARTING
                    self._inner_exception = None

                    header: dict = {
                        "Sec-WebSocket-Protocol": "{}, token-{}".format(self._protocol,
                                                                        self._api_handler.token)
                    }

                    with self._ws_lock:
                        self._ws = WebSocketApp(self._url,
                                                header=header,
                                                on_open=lambda ws: self._check_open(ws),
                                                on_close=lambda ws, code, reason: self._check_close(ws, code, reason),
                                                on_message=lambda ws, msg: self._check_message(ws, msg),
                                                on_error=lambda ws, err: self._check_error(ws, err),
                                                on_ping=lambda ws, data: ws.send(data, opcode=ABNF.OPCODE_PONG))

                    # BUG: While this is initializing, the self._reader_status_condition should remain locked until the
                    # dispatcher listens at the socket. Since we cannot securely release it within, we have to release
                    # it beforehand and expect errors when the websocket gets closed before it was completely
                    # initialized.
                    try:
                        self._reader_guard.release()
                        self._ws.run_forever(http_proxy_host=self._proxy_hostname,
                                             http_proxy_port=self._proxy_port,
                                             http_proxy_auth=self._proxy_auth,
                                             sslopt={
                                                 "cert_reqs": ssl.CERT_NONE
                                             } if AbstractTokenApiHandler.accept_all_certs else None)
                    finally:
                        self._reader_guard.acquire()

                    self._reconnect_delay = self._backoff(self._reconnect_delay)

            except Exception as error:
                self._check_error(self._ws, error)

            self.on_close(self._ws)

            with self._ws_lock:
                self._ws = None

            return self._reader_status, self._inner_exception

    def _backoff(self, reconnect_delay: int) -> int:
        """
        Sleeps for *reconnect_delay* seconds, then returns the delay in seconds for the next try.

        :param reconnect_delay: Delay in seconds to wait.
        :return: Next value for the delay.
        """
        if reconnect_delay:
            with self._reader_guard:
                self._reader_guard.wait(timeout=reconnect_delay)

        return (reconnect_delay + 1) if reconnect_delay < 10 \
            else (reconnect_delay + 10) if reconnect_delay < 60 \
            else random.randint(60, 600)

    def _close(self, status: str = STATUS_NORMAL, reason: str = None):
        """
        Internal stop that does not join the reader thread. Intended to be called on unrecoverable errors within the
        reader thread, i.e. invalid tokens.

        :param status: Status code for close message.
        """
        with self._ws_lock:
            if self._ws:
                self._ws.close(status=status,
                               reason=reason if reason else f"{self._api_handler.user_agent} closing")

    ###############################################################################################################
    # Public API Reader thread
    ###############################################################################################################

    @abstractmethod
    def on_open(self, ws: WebSocketApp):
        pass

    @abstractmethod
    def on_close(self, ws: WebSocketApp, code: int = None, reason: str = None):
        pass

    @abstractmethod
    def on_message(self, ws: WebSocketApp, message: str):
        pass

    @abstractmethod
    def on_error(self, ws: WebSocketApp, error: Exception):
        pass

    ###############################################################################################################
    # Public API Main Writer thread
    ###############################################################################################################

    def start(self) -> None:
        """
        Start the background thread for receiving messages from the websocket. Returns when the websocket has been
        created.

        :raise WebSocketException: When the startup of the reader thread fails.
        """

        try:
            with self._wait_until_finish:
                with self._reader_guard:
                    if self._reader_future:
                        return
                    self._reader_future = self._reader_executor.submit(AbstractAuthenticatedWebSocketHandler._run, self)
                    self._reader_guard.wait()
                    start_ok = (self._reader_status != ReaderStatus.FAILED)

                if not start_ok:
                    self.join()
        except Exception as err:
            raise WebSocketException("Cannot start reader thread.") from err

    def join(self, timeout: int = None) -> None:
        """
        Joins the reader thread and deletes it thereafter.

        :param timeout: Optional timeout in seconds for joining the reader thread.
        :raise WebSocketException: When the reader thread finished with an exception.
        """
        if not self._reader_future:
            return

        try:
            status, exception = self._reader_future.result(timeout)

            if status == ReaderStatus.FAILED:
                if exception:
                    raise exception
                else:
                    raise WebSocketException("Unspecified error")
        except Exception as err:
            raise WebSocketException("Reader thread finished with error.") from err
        finally:
            self._reader_future = None
            with self._wait_until_finish:
                self._wait_until_finish.notify()

    def stop(self, timeout: int = None) -> None:
        """
        Intentionally closes this websocket. Joins on the reader before returning.

        :param timeout: Optional timeout in seconds for joining the reader thread.
        """
        with self._reader_guard:
            self._reader_status = ReaderStatus.DONE
            self._reader_guard.notify()
        self._close()
        self.join(timeout)

    def run_forever(self):
        """
        Runs and receives incoming messages. Will return when the reader thread has been joined. This can happen because
        of an internal error or a call to *stop()* externally.
        """
        with self._wait_until_finish:
            if not self.is_active():
                return
            self._wait_until_finish.wait()

    def is_active(self) -> bool:
        """
        Checks whether the websocket reader thread is still active.
        :return: self._reader_status not in [ReaderStatus.NONE, ReaderStatus.DONE, ReaderStatus.FAILED]
        """
        with self._reader_guard:
            return self._reader_status not in [ReaderStatus.NONE, ReaderStatus.DONE, ReaderStatus.FAILED]

    def restart(self, timeout: int = None) -> bool:
        """
        Closes the websocket and starts a new one. Ensures, that only one thread is attempting to restart.

        :param timeout: Optional timeout in seconds for joining the old reader thread.
        :return: If another thread is already trying to restart, return False, else return True when restart has been
                 executed.
        """
        restart = self._restart_lock.acquire(blocking=False)
        if restart:
            try:
                self.stop(timeout)
                self.start()
            finally:
                self._restart_lock.release()

        return restart

    def send(self, message: str) -> None:
        """
        Send message across the websocket. This is thread-safe.

        :param message: Message as string
        :raise WebSocketException: When *self._auto_reconnect* is False: If a message cannot be sent and all retries
                                   have been exhausted.
        :raise WebSocketConnectionClosedException: When the websocket is not available at all.
        """
        retries = 0
        retry_delay = 0

        while True:
            retry_delay = self._backoff(retry_delay)

            try:
                with self._reader_guard:
                    if self._reader_status in [ReaderStatus.NONE]:
                        raise WebSocketConnectionClosedException('Websocket not started.')
                    elif self._reader_status in [ReaderStatus.DONE, ReaderStatus.FAILED]:
                        raise WebSocketConnectionClosedException('Websocket has exited.')
                    elif self._reader_status not in [ReaderStatus.RUNNING, ReaderStatus.RUNNING_PRELIMINARY]:
                        raise WebSocketException('Websocket not ready.')

                with self._ws_lock:
                    if not self._ws:
                        raise WebSocketConnectionClosedException('Websocket is gone.')

                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug("Sending message: %s", message)
                    else:
                        logger.info("Sending message of length %d", len(message))

                    self._ws.send(message)

                return

            except WebSocketConnectionClosedException:
                raise

            except Exception as err:
                if retries < self.MAX_RETRIES:
                    if self._auto_reconnect:
                        retries = 0
                        logger.warning('Restarting because of error: %s', str(err))
                        self.restart(self._timeout)
                    else:
                        raise WebSocketException("Could not send and all retries have been exhausted.")
                else:
                    logger.warning('Retrying to send message because of error: %s', str(err))
                    retries += 1
